#include <windows.h>
#include <stdio.h>
#include "x86_64_emu.h"
#include "mediana.h"

static u32_t emu_fs_0;

static int win32_int_callback(struct _emu_context *ctx)
{
	printf("win32_int_callback\n");

	return EMU_STATUS_INTERNAL_ERROR;
}

static void *win32_resolve_ip(struct _emu_context *ctx)
{
	return (void*)ctx->regs_gen.EIP;
}

static u8_t win32_loadmem_8(struct _emu_context *ctx, u16_t seg, u64_t offset)
{
	if (seg == 0x3B) {
		return __readfsbyte(u32t(offset));
	}
	return *(u8_t*)offset;
}
static u16_t win32_loadmem16(struct _emu_context *ctx, u16_t seg, u64_t offset)
{
	if (seg == 0x3B) return __readfsword(u32t(offset));
	return *(u16_t*)offset;
}
static u32_t win32_loadmem32(struct _emu_context *ctx, u16_t seg, u64_t offset)
{
	if (seg == 0x3B) {
		if (offset == 0) return emu_fs_0;
		return __readfsdword(u32t(offset));
	}
	return *(u32_t*)offset;
}
static u64_t win32_loadmem64(struct _emu_context *ctx, u16_t seg, u64_t offset)
{
	if (seg == 0x3B) {
		return (u64t(__readfsdword(u32t(offset))) << 32) | __readfsdword(u32t(offset + 4));
	}
	return *(u64_t*)offset;
}
static void win32_savemem_8(struct _emu_context *ctx, u16_t seg, u64_t offset, u8_t data)
{
	if (seg == 0x3B) {
		__writefsbyte(u32t(offset), data);
	} else {
		*(u8_t*)offset = data;
	}
}
static void win32_savemem16(struct _emu_context *ctx, u16_t seg, u64_t offset, u16_t data)
{
	if (seg == 0x3B) {
		__writefsword(u32t(offset), data);
	} else {
		*(u16_t*)offset = data;
	}
}
static void win32_savemem32(struct _emu_context *ctx, u16_t seg, u64_t offset, u32_t data)
{
	if (seg == 0x3B) {
		if (offset == 0) emu_fs_0 = data; return;
		__writefsdword(u32t(offset), data);
	} else {
		*(u32_t*)offset = data;
	}
}
static void win32_savemem64(struct _emu_context *ctx, u16_t seg, u64_t offset, u64_t data)
{
	if (seg == 0x3B) {
		__writefsdword(u32t(offset), data >> 32);
		__writefsdword(u32t(offset + 8), u32t(data));
	} else {
		*(u64_t*)offset = data;
	}
}

static int win32_sysenter(emu_context *ctx, struct INSTRUCTION *instr)
{
	__asm
	{
		pushad
		mov    esi, [ctx]
		mov    eax, emu_context[esi.regs_gen.ESP]
		push   [eax] // esp[0] - virtual return address
		mov    dword ptr [eax], offset after_sysenter
		mov    edi, esp
		mov    eax, emu_context[esi.regs_gen.EAX]
		mov    edx, emu_context[esi.regs_gen.EDX]
		_emit 0x0F
		_emit 0x34
		
after_sysenter:
		mov    esp, edi
		pop    ebx		
		mov    ecx, emu_context[esi.regs_gen.ESP]
		mov    [ecx], ebx
		add    emu_context[esi.regs_gen.ESP], 4
		mov    emu_context[esi.regs_gen.EAX], eax
		mov    emu_context[esi.regs_gen.EIP], ebx		
		popad
	}	
	ctx->jumped = 1; return EMU_STATUS_OK;
}








#define STACK_SIZE (128*1024*1024)


static void test_proc()
{
	MessageBox(0, "test", "test", 0);

	__asm int 3;
}



int main(int argc, char* argv[])
{
	emu_params  params = {0};
	emu_context ctx = {0};
	char       *emu_stack = malloc(STACK_SIZE);
	int         status;

	params.flags = 0;
	params.limit = 0;
	params.mode  = EMU_MODE_32;
	params.on_interrupt = win32_int_callback;
	params.on_resolveip = win32_resolve_ip;
	/* memory callbacks */
	params.do_loadmem_8 = win32_loadmem_8;
	params.do_loadmem16 = win32_loadmem16;
	params.do_loadmem32 = win32_loadmem32;
	params.do_loadmem64 = win32_loadmem64;
	params.do_savemem_8 = win32_savemem_8;
	params.do_savemem16 = win32_savemem16;
	params.do_savemem32 = win32_savemem32;
	params.do_savemem64 = win32_savemem64;
	/* setup SYSENTER handler */
	params.custom_ops[ID_SYSENTER].new_op = win32_sysenter;

	/* setup initial context */
	ctx.regs_seg.CS = 0x1B;
	ctx.regs_seg.DS = 0x23;
	ctx.regs_seg.ES = 0x23;
	ctx.regs_seg.FS = 0x3B;
	ctx.regs_seg.GS = 0;
	ctx.regs_seg.SS = 0x23;
	ctx.regs_gen.EFL = 0x202;
	ctx.regs_gen.EIP = u32t(test_proc);
	ctx.regs_gen.ESP = u32t(emu_stack + STACK_SIZE);

	status = emu_execute_code(&ctx, &params);

	printf("EMU status: %d\n", status);
#ifdef _DEBUG
	Sleep(INFINITE);
#endif
}