#ifndef _EMU_DEFINES_H_
#define _EMU_DEFINES_H_

#include "emu_misc.h"

typedef void (*p_emufun)(emu_context *ctx, struct INSTRUCTION *instr);

/* 8 bit registers */
#define AL ( ctx->regs_gen.AL )
#define AH ( ctx->regs_gen.AH )
#define BL ( ctx->regs_gen.BL )
#define BH ( ctx->regs_gen.BH )
#define CL ( ctx->regs_gen.CL )
#define CH ( ctx->regs_gen.CH )
#define DL ( ctx->regs_gen.DL )
#define DH ( ctx->regs_gen.DH )

/* 16 bit registers */
#define AX ( ctx->regs_gen.AX )
#define BX ( ctx->regs_gen.BX )
#define CX ( ctx->regs_gen.CX )
#define DX ( ctx->regs_gen.DX )
#define SI ( ctx->regs_gen.SI )
#define DI ( ctx->regs_gen.DI )
#define BP ( ctx->regs_gen.BP )
#define SP ( ctx->regs_gen.SP )

/* 32 bit registers */
#define EAX ( ctx->regs_gen.EAX )
#define EBX ( ctx->regs_gen.EBX )
#define ECX ( ctx->regs_gen.ECX )
#define EDX ( ctx->regs_gen.EDX )
#define ESI ( ctx->regs_gen.ESI )
#define EDI ( ctx->regs_gen.EDI )
#define EBP ( ctx->regs_gen.EBP )
#define ESP ( ctx->regs_gen.ESP )

/* 64 bit registers */
#define RAX ( ctx->regs_gen.RAX )
#define RBX ( ctx->regs_gen.RBX )
#define RCX ( ctx->regs_gen.RCX )
#define RDX ( ctx->regs_gen.RDX )
#define RSI ( ctx->regs_gen.RSI )
#define RDI ( ctx->regs_gen.RDI )
#define RBP ( ctx->regs_gen.RBP )
#define RSP ( ctx->regs_gen.RSP )

/* FLAGS */
#define _CF ( ctx->regs_gen.flags._cf )
#define _PF ( ctx->regs_gen.flags._pf )
#define _AF ( ctx->regs_gen.flags._af )
#define _ZF ( ctx->regs_gen.flags._zf )
#define _SF ( ctx->regs_gen.flags._sf )
#define _TF ( ctx->regs_gen.flags._tf )
#define _IF ( ctx->regs_gen.flags._if )
#define _DF ( ctx->regs_gen.flags._df )
#define _OF ( ctx->regs_gen.flags._of )
#define EFL ( ctx->regs_gen.EFL )
#define RFL ( ctx->regs_gen.RFL )

/* segment registers */
#define CS ( ctx->regs_seg.CS )
#define DS ( ctx->regs_seg.DS )
#define ES ( ctx->regs_seg.ES )
#define SS ( ctx->regs_seg.SS )
#define FS ( ctx->regs_seg.FS )
#define GS ( ctx->regs_seg.GS )

#define CPL  ( CS & 3 )
#define IOPL ( ctx->regs_gen.flags._iopl )

/* instruction pointer */
#define RIP ( ctx->regs_gen.RIP )
#define EIP ( ctx->regs_gen.EIP )

#define clearCF() _CF = 0
#define clearPF() _PF = 0
#define clearAF() _AF = 0
#define clearZF() _ZF = 0
#define clearSF() _SF = 0
#define clearOF() _OF = 0
#define clearIF() _IF = 0
#define clearDF() _DF = 0

#define setCF() _CF = 1
#define setPF() _PF = 1
#define setAF() _AF = 1
#define setZF() _ZF = 1
#define setSF() _SF = 1
#define setOF() _OF = 1
#define setIF() _IF = 1
#define setDF() _DF = 1

#define FLAGS_CF_MASK   (1 <<  0)
#define FLAGS_PF_MASK   (1 <<  2)
#define FLAGS_AF_MASK   (1 <<  4)
#define FLAGS_ZF_MASK   (1 <<  6)
#define FLAGS_SF_MASK   (1 <<  7)
#define FLAGS_TF_MASK   (1 <<  8)
#define FLAGS_IF_MASK   (1 <<  9)
#define FLAGS_DF_MASK   (1 << 10)
#define FLAGS_OF_MASK   (1 << 11)
#define FLAGS_IOPL_MASK (3 << 12)
#define FLAGS_NT_MASK   (1 << 14)
#define FLAGS_RF_MASK   (1 << 16)
#define FLAGS_VM_MASK   (1 << 17)
#define FLAGS_AC_MASK   (1 << 18)
#define FLAGS_VIF_MASK  (1 << 19)
#define FLAGS_VIP_MASK  (1 << 20)
#define FLAGS_ID_MASK   (1 << 21)

#define FLAGS_OSZAPC_MASK \
    (FLAGS_CF_MASK | FLAGS_PF_MASK | FLAGS_AF_MASK | FLAGS_ZF_MASK | FLAGS_SF_MASK | FLAGS_OF_MASK)

#define FLAGS_OSZAPMask  \
    (FLAGS_PF_MASK | FLAGS_AF_MASK | FLAGS_ZF_MASK | FLAGS_SF_MASK | FLAGS_OF_MASK)

#define LAST_BIT(_bits) \
	(1ULL << ((_bits) - 1))

#define BIT_MASK(_bits) \
	((_bits) == 64 ? 0xFFFFFFFFFFFFFFFF : (1ULL << ((_bits) & 0x3F)) - 1)

#define S_EXTD(_val, _from_bits, _to_bits) \
	(s##_to_bits##t(s##_from_bits##t(_val)))

#define BSWAP_16(_op) ( \
	((_op >> 8) & 0x00FF) | ((_op << 8) & 0xFF00) )

#define BSWAP_32(_op) (                                       \
	((_op >> 24) & 0x000000FF) | ((_op >> 8)  & 0x0000FF00) | \
	((_op << 8)  & 0x00FF0000) | ((_op << 24) & 0xFF000000) )

#define BSWAP_64(_op) (                                                     \
	((_op >> 56) & 0x000000000000FF) | ((_op >> 40) & 0x000000000000FF00) | \
	((_op >> 24) & 0x00000000FF0000) | ((_op >> 8)  & 0x00000000FF000000) | \
	((_op << 8)  & 0x0000FF00000000) | ((_op << 24) & 0x0000FF0000000000) | \
	((_op << 40) & 0xFF000000000000) | ((_op << 56) & 0xFF00000000000000) )

#define OP0_SIZE u32t( instr->ops[0].size )
#define OP0_BITS u32t( instr->ops[0].size * 8 )
#define OP1_SIZE u32t( instr->ops[1].size )
#define OP1_BITS u32t( instr->ops[1].size * 8 )

#define OP0_REG_CODE ( instr->ops[0].value.reg.code )
#define OP1_REG_CODE ( instr->ops[1].value.reg.code )

#define bit_op_mask(_op) \
	((instr->ops[0].size * 8) - 1)

#define bit_op_bit(_op) \
	((_op) & bit_op_mask(_op))

#define bit_op_off(_op) \
	( ((instr->ops[0].flags & OPERAND_TYPE_MEM) && (instr->ops[1].flags & OPERAND_TYPE_REG)) ? \
	  ((op1 & ~bit_op_mask(_op)) / 8) : 0 )

#define EMU_DO_DIV(_src, _HI, _LO, _dtype, _bits) {             \
	_dtype _op1 = ((_dtype)(_HI) << (_bits)) | ((_dtype)(_LO)); \
	_dtype _tmp = _op1 / ((_src) & BIT_MASK(_bits));            \
                                                                \
	if (_tmp > BIT_MASK(_bits)) {                               \
		exception(EMU_EXCEPT_DE, 0);                            \
	}                                                           \
	_LO = _tmp & BIT_MASK(_bits);                               \
	_HI = _op1 % ((_src) & BIT_MASK(_bits));                    \
}

#define EMU_DO_IDIV(_src, _HI, _LO, _type, _dtype, _bits) {                            \
	_dtype _op1 = ((_dtype)(_HI) << (_bits)) | ((_dtype)(_LO));                        \
	_dtype _tmp = _op1 / ((_type)(_src));                                              \
                                                                                       \
	if (_tmp > (_dtype)(LAST_BIT(_bits) - 1) || _tmp < -((_dtype)(LAST_BIT(_bits)))) { \
		exception(EMU_EXCEPT_DE, 0);                                                   \
	}                                                                                  \
	_LO = (_type)(_tmp);                                                               \
	_HI = _op1 % ((_type)_src);                                                        \
}

#define EMU_DO_IMUL(_res, _op1, _op2, _bits, _dbits) {                              \
	s##_dbits##_t _tmp = S_EXTD(_op1, _bits, _dbits) * S_EXTD(_op2, _bits, _dbits); \
	u##_bits##_t  _tlo = u##_bits##t(_tmp);                                         \
                                                                                    \
	_CF = _OF = (_tmp != s##_bits##t(_tmp));                                        \
	_SF = _ZF = _AF = _PF = 0;                                                      \
	                                                                                \
	_res = _tmp;                                                                    \
}

#define ENU_DO_MUL(_LO, _HI, _op2, _bits, _dbits) {              \
	u##_dbits##_t _res = u##_dbits##t(_LO) * u##_dbits##t(_op2); \
	u##_bits##_t  _rhi = _res >> (_bits);                        \
	u##_bits##_t  _rlo = u##_bits##t(_res);                      \
	                                                             \
    _LO = _rlo;                                                  \
	_HI = _rhi;                                                  \
	                                                             \
	_CF = _OF = (_rhi != 0);                                     \
	_SF = _ZF = _AF = _PF = 0;                                   \
}

#define EMU_DO_ENTER(_SP, _BP, _level, _imm16, _bits, _opsize) {                  \
	u##_bits##_t _tmpSP, _tmpBP = _BP;                                            \
                                                                                  \
    emu_push(ctx, _tmpBP, _opsize);                                               \
    _tmpSP = _SP;                                                                 \
                                                                                  \
	if (_level > 0)                                                               \
	{                                                                             \
		/* do level-1 times */                                                    \
		while (--_level) {						                                  \
			_tmpBP   -= _opsize;                                                  \
			emu_push(ctx, read_mem_size(SREG_CODE_SS, _tmpBP, _opsize), _opsize); \
		}                                                                         \
		/* push(frame pointer) */                                                 \
		emu_push(ctx, _tmpSP, _opsize);                                           \
	}                                                                             \
	_SP -= _imm16;                                                                \
                                                                                  \
    /* ENTER finishes with memory write check on the final stack  */              \
	/* pointer the memory is touched but no write actually occurs */              \
	/* emulate it by doing RMW read access from SS:SP             */              \
	read_mem_size(SREG_CODE_SS, _SP, _opsize);                                    \
	_BP = _tmpSP;                                                                 \
}

#define EMU_DO_LEAVE(_SP, _BP, _bits) {                                   \
	u##_bits##_t _temp = u##_bits##t(read_mem##_bits(SREG_CODE_SS, _BP)); \
	                                                                      \
	_SP = _BP + (_bits / 8);                                              \
	_BP = _temp;                                                          \
}

#define ROT_MASK \
	((OP0_SIZE == OPERAND_SIZE_64) ? 0x3F : 0x1F)

#define SET_FLAGS_SZP(_res, _size) {      \
	_SF = (_res & sign_bits[_size]) != 0; \
	_ZF = (_res & mask_bits[_size]) == 0; \
	_PF = parity_table[_res & 0xFF];      \
}

#define SET_FLAGS_OSZAPC_LOGIC(_res, _size) { \
	_CF = 0; _AF = 0; _OF = 0;                \
	SET_FLAGS_SZP(_res, _size);               \
}

#define SET_FLAGS_OSZAPC_INC(_res, _size) {              \
	/* CF flag is not affected */                        \
	_AF = (_res & 0xF) == 0;                             \
	_OF = (_res & mask_bits[_size]) == sign_bits[_size]; \
	SET_FLAGS_SZP(_res, _size);                          \
}

#define SET_FLAGS_OSZAPC_DEC(_res, _size) {                    \
	/* CF flag is not affected */                              \
	_AF = (_res & 0xF) == 0xF;                                 \
	_OF = (_res & mask_bits[_size]) == (sign_bits[_size] - 1); \
	SET_FLAGS_SZP(_res, _size);                                \
}

#define SET_FLAGS_OSZAPC_NEG(_res, _size) {              \
	_CF = (_res & mask_bits[_size]) != 0;                \
	_AF = (_res & 0xF) != 0;                             \
	_OF = (_res & mask_bits[_size]) == sign_bits[_size]; \
	SET_FLAGS_SZP(_res, _size);                          \
}

#define SET_FLAGS_OSZAPC_ADD(_op1, _op2, _sum, _size) {                             \
	u64_t _sign = sign_bits[_size];                                                 \
	u64_t _mask = mask_bits[_size];                                                 \
	_CF = (_sum & _mask) < (_op1 & _mask);                                          \
	_AF = ((_op1 ^ _op2 ^ _sum) & 0x10) != 0;                                       \
	_OF = ((_op1) & (_op2) & ~(_sum) & _sign) || (~(_op1) & ~(_op2) & sum & _sign); \
	SET_FLAGS_SZP(_sum, _size);                                                     \
}

#define SET_FLAGS_OSZAPC_ADC(_op1, _op2, _sum, _size) {                                  \
	u64_t _sign = sign_bits[_size];                                                      \
	u64_t _mask = mask_bits[_size];                                                      \
	_CF = _CF == 0 ? ((_sum & _mask) <  (_op1 & _mask)) :                                \
                     ((_sum & _mask) <= (_op1 & _mask));                                 \
	_AF = ((_op1 ^ _op2 ^ _sum) & 0x10) != 0;                                            \
	_OF = ((_op1) & (_op2) & ~(_sum) & _sign) || (~(_op1) & ~(_op2) & sum & _sign);      \
	SET_FLAGS_SZP(_sum, _size);                                                          \
}

#define SET_FLAGS_OSZAPC_SUB(_op1, _op2, _dif, _size) {                                \
	u64_t _sign = sign_bits[_size];                                                    \
	u64_t _mask = mask_bits[_size];                                                    \
	_CF = (_op1 & _mask) < (_op2 & _mask);                                             \
	_AF = ((_op1 ^ _op2 ^ _dif) & 0x10) != 0;                                          \
	_OF = ((_op1) & ~(_op2) & ~(_dif) & _sign) || (~(_op1) & (_op2) & (_dif) & _sign); \
	SET_FLAGS_SZP(_dif, _size);                                                        \
}

#define SET_FLAGS_OSZAPC_SBB(_op1, _op2, _dif, _size) {                                \
	u64_t _sign = sign_bits[_size];                                                    \
	u64_t _mask = mask_bits[_size];                                                    \
	_CF = _CF == 0 ? ( (_op1 & _mask) < (_op2 & _mask)) :                              \
	                 (((_op1 & _mask) < (_dif & _mask)) || ((_op2 & _mask) == _mask)); \
    _AF = ((_op1 ^ _op2 ^ _dif) & 0x10) != 0;                                          \
	_OF = ((_op1) & ~(_op2) & ~(_dif) & _sign) || (~(_op1) & (_op2) & (_dif) & _sign); \
	SET_FLAGS_SZP(_dif, _size);                                                        \
}

#endif