/*
   getted from BOCHS sources
*/
#include "x86_64_emu.h"
#include "emu_math.h"

static unsigned partial_add(u32_t *sum, u32_t b)
{
  u32_t t = *sum;
  *sum += b;
  return (*sum < t);
}

void long_mul(u128_t *product, u64_t op1, u64_t op2)
{
  u32_t op_1[2],op_2[2];
  u32_t result[5];
  u64_t nn;
  unsigned c;

  int i,j,k;

  op_1[0] = u32t(op1 & 0xffffffff);
  op_1[1] = u32t(op1 >> 32);
  op_2[0] = u32t(op2 & 0xffffffff);
  op_2[1] = u32t(op2 >> 32);

  for (i = 0; i < 4; i++) result[i] = 0;

  for (i = 0; i < 2; i++) {
    for (j = 0; j < 2; j++) {
      nn = u64t(op_1[i]) * u64t(op_2[j]);
      k = i + j;
      c = partial_add(&result[k++], u32t(nn & 0xffffffff));
      c = partial_add(&result[k++], u32t(nn >> 32) + c);
      while (k < 4 && c != 0) {
        c = partial_add(&result[k++], c);
      }
    }
  }

  product->lo = result[0] + (u64t(result[1]) << 32);
  product->hi = result[2] + (u64t(result[3]) << 32);
}

void long_neg(s128_t *n)
{
  u64_t t = n->lo;
  n->lo = - s64t(n->lo);
  if (t - 1 > t) --n->hi;
  n->hi = ~n->hi;
}

void long_imul(s128_t *product, s64_t op1, s64_t op2)
{
  unsigned s1,s2;

  if ((s1 = (op1 < 0))) op1 = -op1;
  if ((s2 = (op2 < 0))) op2 = -op2;
  long_mul((u128_t*)product,u64t(op1),u64t(op2));
  if (s1 ^ s2)
    long_neg(product);
}

void long_shl(u128_t *a)
{
  u64_t c;
  c = a->lo >> 63;
  a->lo <<= 1;
  a->hi <<= 1;
  a->hi |= c;
}

void long_shr(u128_t *a)
{
  u64_t c;
  c = a->hi << 63;
  a->hi >>= 1;
  a->lo >>= 1;
  a->lo |= c;
}

unsigned long_sub(u128_t *a, u128_t *b)
{
  int c;
  u64_t t = a->lo;
  a->lo -= b->lo;
  c = (a->lo > t);
  t = a -> hi;
  a->hi -= b->hi + c;
  return(a->hi > t);
}

int long_le(u128_t *a, u128_t *b)
{
  if (a->hi == b->hi) {
    return(a->lo <= b->lo);
  } else {
    return(a->hi <= b->hi);
  }
}

void long_div(u128_t *quotient, u64_t *remainder, u128_t *dividend, u64_t divisor)
{
  /*
  n := 0;
  while (divisor <= dividend) do
    inc(n);
    divisor := divisor * 2;
  end;
  quotient := 0;
  while n > 0 do
    divisor := divisor div 2;
    quotient := quotient * 2;
    temp := dividend;
    dividend := dividend - divisor;
    if temp > dividend then
      dividend := temp;
    else
      inc(quotient);
    end;
    dec(n);
  end;
  remainder := dividend;
  */

  u128_t d,acc,q,temp;
  int n,c;

  d.lo = divisor;
  d.hi = 0;
  acc.lo = dividend->lo;
  acc.hi = dividend->hi;
  q.lo = 0;
  q.hi = 0;
  n = 0;

  while (long_le(&d,&acc) && n < 128) {
    long_shl(&d);
    n++;
  }

  while (n > 0) {
    long_shr(&d);
    long_shl(&q);
    temp.lo = acc.lo;
    temp.hi = acc.hi;
    c = long_sub(&acc,&d);
    if (c) {
      acc.lo = temp.lo;
      acc.hi = temp.hi;
    } else {
      q.lo++;
    }
    n--;
  }

  *remainder = acc.lo;
  quotient->lo  = q.lo;
  quotient->hi  = q.hi;
} 

void long_idiv(s128_t *quotient, s64_t *remainder, s128_t *dividend, s64_t divisor)
{
  unsigned s1,s2;
  s128_t temp;

  temp = *dividend;
  if ((s1 = (temp.hi < 0))) {
    long_neg(&temp);
  }
  if ((s2 = (divisor < 0))) divisor = -divisor;
  long_div((u128_t*)quotient,(u64_t*)remainder,(u128_t*)&temp,divisor);
  if (s1 ^ s2) {
    long_neg(quotient);
  }
  if (s2) {
    *remainder = -*remainder;
  }
}