#ifndef _EMU_MISC_H_
#define _EMU_MISC_H_

u64_t emu_get_address(emu_context *ctx, struct INSTRUCTION *instr, int op_n, int *sreg);
u64_t emu_load_operand(emu_context *ctx, struct INSTRUCTION *instr, int op_n, u64_t mem_off);
void  emu_save_operand(emu_context *ctx, struct INSTRUCTION *instr, u64_t data, int op_n, u64_t mem_off);

void  emu_push(emu_context *ctx, u64_t data, u16_t opsize);
u64_t emu_pop(emu_context *ctx, u16_t opsize);

#define load_operand(_op_n, _mem_off) \
	emu_load_operand(ctx, instr, _op_n, _mem_off)

#define save_operand(_op_n, _val, _mem_off) \
	emu_save_operand(ctx, instr, _val, _op_n, _mem_off)


#define load_imm_op(_op_n) ( \
	(instr->ops[_op_n].size == OPERAND_SIZE_8)  ? instr->ops[_op_n].value.imm.imm8  : \
	(instr->ops[_op_n].size == OPERAND_SIZE_16) ? instr->ops[_op_n].value.imm.imm16 : \
	(instr->ops[_op_n].size == OPERAND_SIZE_32) ? instr->ops[_op_n].value.imm.imm32 : \
	                                              instr->ops[_op_n].value.imm.imm64 )

#define get_address(_op_n, _seg) \
	emu_get_address(ctx, instr, _op_n, _seg)

#define exception(_vector, _error_code) { \
	ctx->params->int_num = _vector;       \
	ctx->params->int_arg = _error_code;   \
	                                      \
	longjmp(ctx->params->saved_env, 1);   \
}

#define MASK_MODE(_offset) \
	((_offset) & mode_mask[ctx->params->mode])

#define MASK_ADDR(_offset) \
	((_offset) & mask_bits[instr->addrsize] & mode_mask[ctx->params->mode])

#define read_mem8(_sreg, _offset) \
	ctx->params->do_loadmem_8(ctx, get_regseg(_sreg), MASK_ADDR(_offset))

#define read_mem16(_sreg, _offset) \
	ctx->params->do_loadmem16(ctx, get_regseg(_sreg), MASK_ADDR(_offset))

#define read_mem32(_sreg, _offset) \
	ctx->params->do_loadmem32(ctx, get_regseg(_sreg), MASK_ADDR(_offset))

#define read_mem64(_sreg, _offset) \
	ctx->params->do_loadmem64(ctx, get_regseg(_sreg), MASK_ADDR(_offset))

#define write_mem8(_sreg, _offset, _data) \
	ctx->params->do_savemem_8(ctx, get_regseg(_sreg), MASK_ADDR(_offset), u8t(_data))

#define write_mem16(_sreg, _offset, _data) \
	ctx->params->do_savemem16(ctx, get_regseg(_sreg), MASK_ADDR(_offset), u16t(_data))

#define write_mem32(_sreg, _offset, _data) \
	ctx->params->do_savemem32(ctx, get_regseg(_sreg), MASK_ADDR(_offset), u32t(_data))

#define write_mem64(_sreg, _offset, _data) \
	ctx->params->do_savemem64(ctx, get_regseg(_sreg), MASK_ADDR(_offset), u64t(_data))

#define read_mem_size(_sreg, _offset, _size) ( \
	(_size == OPERAND_SIZE_8)  ? read_mem8(_sreg, _offset) : \
	(_size == OPERAND_SIZE_16) ? read_mem16(_sreg, _offset) : \
	(_size == OPERAND_SIZE_32) ? read_mem32(_sreg, _offset) : read_mem64(_sreg, _offset) )


#define write_mem_size(_sreg, _offset, _data, _size) {                     \
	if (_size == OPERAND_SIZE_8)  write_mem8(_sreg, _offset, _data);  else \
	if (_size == OPERAND_SIZE_16) write_mem16(_sreg, _offset, _data); else \
	if (_size == OPERAND_SIZE_32) write_mem32(_sreg, _offset, _data); else \
	                              write_mem64(_sreg, _offset, _data);      \
}
	 
#define get_reggen64(_reg) ( ctx->regs_gen.regs[_reg].reg64  )
#define get_reggen32(_reg) ( ctx->regs_gen.regs[_reg].reg32  )
#define get_reggen16(_reg) ( ctx->regs_gen.regs[_reg].reg16  )
#define get_reggen8(_reg)  ( ((u8_t*)ctx)[reg8_offset[_reg]] )
#define get_regseg(_reg)   ( ctx->regs_seg.regs[_reg]        )

#define get_reggen(_reg, _size) (                     \
	(_size) == OPERAND_SIZE_8  ? get_reggen8(_reg) :  \
	(_size) == OPERAND_SIZE_16 ? get_reggen16(_reg) : \
	(_size) == OPERAND_SIZE_32 ? get_reggen32(_reg) : get_reggen64(_reg) )

#define get_reg_addr(_reg) \
	( ctx->params->mode == EMU_MODE_16 ? get_reggen16(_reg) : \
	  ctx->params->mode == EMU_MODE_32 ? get_reggen32(_reg) : get_reggen64(_reg) )

#define set_reg_addr(_reg, _val)                                 \
	switch(ctx->params->mode) {                                  \
		case EMU_MODE_16: set_reggen16(_reg, u16t(_val)); break; \
		case EMU_MODE_32: set_reggen32(_reg, u32t(_val)); break; \
		default: set_reggen64(_reg, u64t(_val)); break;          \
	}

#define get_regseg(_reg) ( ctx->regs_seg.regs[_reg] )

#define set_reggen64(_reg, _val) ( ctx->regs_gen.regs[_reg].reg64 = u64t(_val) )
#define set_reggen32(_reg, _val) ( ctx->regs_gen.regs[_reg].reg32 = u32t(_val) )
#define set_reggen16(_reg, _val) ( ctx->regs_gen.regs[_reg].reg16 = u16t(_val) )
#define set_reggen8(_reg, _val)  ( ((u8_t*)ctx)[reg8_offset[_reg]] = u8t(_val) )

#define set_reggen(_reg, _val, _size) {                              \
	switch(_size) {                                                  \
		case OPERAND_SIZE_8:  set_reggen8(_reg, u8t(_val)); break;   \
		case OPERAND_SIZE_16: set_reggen16(_reg, u16t(_val)); break; \
		case OPERAND_SIZE_32: set_reggen32(_reg, u32t(_val)); break; \
		default: set_reggen64(_reg, u64t(_val)); break;              \
	}                                                                \
}

#define push16(_data) emu_push(ctx, _data, OPERAND_SIZE_16)
#define push32(_data) emu_push(ctx, _data, OPERAND_SIZE_32)
#define push64(_data) emu_push(ctx, _data, OPERAND_SIZE_64)

#define pop16() ( u16t(emu_pop(ctx, OPERAND_SIZE_16)) )
#define pop32() ( u32t(emu_pop(ctx, OPERAND_SIZE_32)) )
#define pop64() ( u64t(emu_pop(ctx, OPERAND_SIZE_64)) )


extern const u64_t sign_bits[];
extern const u64_t mask_bits[];
extern const u64_t mode_mask[];
extern const u8_t  parity_table[256];
extern const u32_t reg8_offset[20];

#endif